from typing import Any, Callable

Schema: Any
Table: Any
Field: Any
DataType: Any
schema: Any
map_: Any
list_: Any
field: Any
struct: Any
type_for_alias: Any
date32: Any
date64: Any
decimal128: Any
float16: Any
float32: Any
float64: Any

py_buffer: Callable[[bytes], Any]
NativeFile: Any
BufferReader: Any
